/*
 * IP packet frag compiler.
 *	04/23/2002
 *
 * skape
 * mmiller@hick.org
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <net/ethernet.h>
#define __FAVOR_BSD
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>

#include <pcap.h>

#define NUM_HASH_BUCKETS	256
#define MAX_DATA_LEN            16384

typedef struct _packet {

	unsigned long  recvTime;

	unsigned char  *pkt;
	unsigned long  len;

	struct _packet *prev;
	struct _packet *next;

} PACKET;

typedef struct _frag_packet {

	struct ip     ip;

	union {
		struct tcphdr  tcp;
		struct udphdr  udp;
		struct icmp icmp;
	} proto;

	unsigned char *data;
	unsigned long dataLen;

} FRAG_PACKET;

typedef struct _fragcomp_entry {
	
	unsigned char id[11];	/* proto=1 id=2 src=4 dst=4 */
	FRAG_PACKET   *fragPacket;

	struct _fragcomp_entry *prev;
	struct _fragcomp_entry *next;

} FRAGCOMP_ENTRY;

typedef struct _fragcomp_global {

	const char      *device;

	FRAGCOMP_ENTRY *fragHash[NUM_HASH_BUCKETS];

} FRAGCOMP_GLOBAL;

void packetReceive();

void packetHashGetId(unsigned char *ipPacket, unsigned char *id);
FRAG_PACKET *packetHashLookup(unsigned char *ipPacket);
FRAG_PACKET *packetHashPacket(unsigned char *ipPacket, unsigned long ipPacketLen);
FRAG_PACKET *packetHashRemove(unsigned char *ipPacket);

FRAGCOMP_GLOBAL global;

int main(int argc, char **argv)
{
	int c;

	memset(&global,0,sizeof(global));

	/* Initialization */

	global.device      = "eth0";

	while ((c = getopt(argc, argv, "i:")) != EOF)
	{
		switch (c)
		{
			case 'i':
				global.device = optarg;
				break;
		}
	}

	fprintf(stdout,"Listening for fragmented packets on '%s'...\n",global.device);

	packetReceive();

	return 1;
}

void packetReceive()
{
	struct pcap_pkthdr pchdr;
	unsigned char      *newPacket;
	char               errorBuffer[PCAP_ERRBUF_SIZE];
	pcap_t             *pcapFd;
	struct ip          *ip;

	if ((pcapFd = pcap_open_live((char *)global.device, 128, 0, 0, errorBuffer)) == NULL)
	{
		fprintf(stderr,"pcap_open_live(): %s\n", errorBuffer);
		return;
	}

	while ((newPacket = (char *)pcap_next(pcapFd, &pchdr)) != NULL)
	{
		if ((pchdr.len < sizeof(struct ether_header)) || 
			(((struct ether_header *)newPacket)->ether_type != htons(ETHERTYPE_IP)))
			continue;

		ip = (struct ip *)((unsigned char *)newPacket + sizeof(struct ether_header));

		/* Check to see if more frags are incoming */

		if (((unsigned char *)&ip->ip_off)[0] & 0x20)
			packetHashPacket((unsigned char *)ip, pchdr.len - sizeof(struct ether_header));
		else
		{
			/* Check to see if this id for this protocol had frags and close them out */
			FRAG_PACKET *fcPacket;

			/* If this has a fragment offset... */
			if (((unsigned char *)&ip->ip_off)[1])
				packetHashPacket((unsigned char *)ip, pchdr.len - sizeof(struct ether_header));

			if ((fcPacket = packetHashLookup((unsigned char *)ip)))
			{
				fcPacket = packetHashRemove((unsigned char *)ip);

				if (fcPacket)
				{
					int x = 0;
					fprintf(stdout,"------------------------------------------------------------------------------\n");
					fprintf(stdout,"Status     : Fragged packet compilation done for id=%04x proto=%s\n",
								fcPacket->ip.ip_id,
								(fcPacket->ip.ip_p == IPPROTO_TCP)?"TCP":
									(fcPacket->ip.ip_p == IPPROTO_UDP)?"UDP":
									(fcPacket->ip.ip_p == IPPROTO_ICMP)?"ICMP":"UNKNOWN");
					fprintf(stdout,"Src        : %hi.%hi.%hi.%hi\n",
								((unsigned char *)&fcPacket->ip.ip_src.s_addr)[0],
								((unsigned char *)&fcPacket->ip.ip_src.s_addr)[1],
								((unsigned char *)&fcPacket->ip.ip_src.s_addr)[2],
								((unsigned char *)&fcPacket->ip.ip_src.s_addr)[3]);
					fprintf(stdout,"Dst        : %hi.%hi.%hi.%hi\n",
								((unsigned char *)&fcPacket->ip.ip_dst.s_addr)[0],
								((unsigned char *)&fcPacket->ip.ip_dst.s_addr)[1],
								((unsigned char *)&fcPacket->ip.ip_dst.s_addr)[2],
								((unsigned char *)&fcPacket->ip.ip_dst.s_addr)[3]);

					switch(fcPacket->ip.ip_p)
					{
						case IPPROTO_TCP:
							fprintf(stdout,"Src Port   : %d\n",fcPacket->proto.tcp.th_sport);
							fprintf(stdout,"Dst Port   : %d\n",fcPacket->proto.tcp.th_dport);
							break;
						case IPPROTO_UDP:
							fprintf(stdout,"Src Port   : %d\n",fcPacket->proto.udp.uh_sport);
							fprintf(stdout,"Dst Port   : %d\n",fcPacket->proto.udp.uh_dport);
							break;
					}

					fprintf(stdout,"Data Length: %lu\n",fcPacket->dataLen);

					if (fcPacket->data)
					{
						fprintf(stdout,"Data       : ");
						for (x = 0; x < fcPacket->dataLen; x++)
							fprintf(stdout,"%02x",fcPacket->data[x] & 0xFF);
	
						fprintf(stdout,"\n");

						free(fcPacket->data);
					}

					free(fcPacket);
				}
			}
		}
	}
}

void packetHashGetId(unsigned char *ipPacket, unsigned char *id)
{
	struct ip *ip = (struct ip *)ipPacket;

	/* proto=1 id=2 src=4 dst=4 */
	memset(id, 0, 11);

	memcpy(id,&ip->ip_p,1);		  /* Protocol 		*/
	memcpy(id+1,&ip->ip_id,2);	  /* Id			*/
	memcpy(id+3,&ip->ip_src.s_addr,4);/* src		*/
	memcpy(id+7,&ip->ip_dst.s_addr,4);/* dst		*/
}

FRAG_PACKET *packetHashLookup(unsigned char *ipPacket)
{
	unsigned char id[11];
	unsigned long *idPos = (unsigned long *)id, realPos;
	FRAGCOMP_ENTRY *entry = NULL;

	packetHashGetId(ipPacket, id);

	realPos = (*idPos & (NUM_HASH_BUCKETS - 1));

	for (((entry = global.fragHash[realPos]));
		entry;
		entry = entry->next)
	{
		if (memcmp(entry->id,id,11) == 0)
			break;	
	}

	return (entry)?entry->fragPacket:NULL;
}

FRAG_PACKET *packetHashPacket(unsigned char *ipPacket, unsigned long ipPacketLen)
{
	unsigned char id[11];
	unsigned long *idPos = (unsigned long *)id, realPos;
	FRAGCOMP_ENTRY *entry = NULL;
	unsigned char *dataBuf = NULL;
	unsigned long dataLen = 0;

	packetHashGetId(ipPacket, id);

	realPos = (*idPos & (NUM_HASH_BUCKETS - 1));

	for (((entry = global.fragHash[realPos]));
		entry;
		entry = entry->next)
	{
		if (memcmp(entry->id,id,11) == 0)
			break;	
	}

	if (!entry)
	{
		unsigned int headerLength;

		entry = (FRAGCOMP_ENTRY *)malloc(sizeof(FRAGCOMP_ENTRY));
		memset(entry,0,sizeof(FRAGCOMP_ENTRY));

		memcpy(entry->id, id, 11);

		entry->fragPacket = (FRAG_PACKET *)malloc(sizeof(FRAG_PACKET));
		memset(entry->fragPacket, 0, sizeof(FRAG_PACKET));

		memcpy(&entry->fragPacket->ip, ipPacket, sizeof(entry->fragPacket->ip));

		headerLength = entry->fragPacket->ip.ip_hl * 4;

		switch(entry->fragPacket->ip.ip_p)
		{
			case IPPROTO_TCP:
				if (headerLength + sizeof(struct tcphdr) <= ipPacketLen)
				{
					memcpy(&entry->fragPacket->proto.tcp, ipPacket + headerLength, sizeof(struct tcphdr));
					dataBuf = ipPacket + headerLength + sizeof(struct tcphdr);
					dataLen = ipPacketLen - (headerLength + sizeof(struct tcphdr));
				}
				break;
			case IPPROTO_UDP:
				if (headerLength + sizeof(struct udphdr) <= ipPacketLen)
				{
					memcpy(&entry->fragPacket->proto.udp, ipPacket + headerLength, sizeof(struct udphdr));
					dataBuf = ipPacket + headerLength + sizeof(struct udphdr);
					dataLen = ipPacketLen - (headerLength + sizeof(struct udphdr));
				}
				break;
			case IPPROTO_ICMP:
				if (headerLength + 8 <= ipPacketLen)
				{
					memcpy(&entry->fragPacket->proto.icmp, ipPacket + headerLength, 8);
					dataBuf = ipPacket + headerLength + 8;
					dataLen = ipPacketLen - (headerLength + 8);
				}
				break;
		}

		entry->next = global.fragHash[realPos];

		if (entry->next)
			entry->next->prev = entry;

		global.fragHash[realPos] = entry;
	}
	else
	{
		struct ip *ip = (struct ip *)ipPacket;

		dataBuf = (unsigned char *)ipPacket+(ip->ip_hl*4);
		dataLen = ipPacketLen - (ip->ip_hl*4);
	}

	if ((dataLen > 0) && (entry->fragPacket->data))
	{
		if (entry->fragPacket->dataLen + ipPacketLen > MAX_DATA_LEN)
			fprintf(stdout,"packet exceeds data size....\n");
		else
		{
			entry->fragPacket->data = realloc(entry->fragPacket->data, entry->fragPacket->dataLen + ipPacketLen);
			memcpy(entry->fragPacket->data + entry->fragPacket->dataLen, dataBuf, dataLen);
			entry->fragPacket->dataLen += dataLen;
		}
	}
	else if (dataLen > 0)
	{
		entry->fragPacket->data = (unsigned char *)malloc(dataLen);
		memcpy(entry->fragPacket->data,dataBuf,(entry->fragPacket->dataLen = dataLen));
	}

	return (entry)?entry->fragPacket:NULL;
}

FRAG_PACKET *packetHashRemove(unsigned char *ipPacket)
{
	unsigned char id[11];
	unsigned long *idPos = (unsigned long *)id, realPos;
	FRAGCOMP_ENTRY *entry = NULL;
	FRAG_PACKET *ret = NULL;

	packetHashGetId(ipPacket, id);

	realPos = (*idPos & (NUM_HASH_BUCKETS - 1));

	for (((entry = global.fragHash[realPos]));
		entry;
		entry = entry->next)
	{
		if (memcmp(entry->id,id,11) == 0)
			break;	
	}

	if (entry)
	{
		if (entry == global.fragHash[realPos])
			global.fragHash[realPos] = entry->next;

		if (entry->prev)
			entry->prev->next = entry->next;
		if (entry->next)
			entry->next->prev = entry->prev;

		ret = entry->fragPacket;

		free(entry);
	}

	return ret;
}

